package laboratoire4;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

public class MiniMax {
    private static final int MAX_MAX_DEPTH = 6;
    private static int MAX_DEPTH = 2;

    public static MiniMaxResult miniMax(PusherBoard board) {
        long startMillis = System.currentTimeMillis();
        MovingBoard game = new MovingBoard(board.getBoard(), board.getPlayer());
        Collection<Action> actions = getActions(game, true);

        MiniMaxResult maxResult = null;

        for (int i = 0; i <= MAX_MAX_DEPTH; i += 2) {
            MAX_DEPTH = i;
            for (Action action : actions) {
                int score = min(game, 0, Integer.MIN_VALUE, Integer.MAX_VALUE);

                if (maxResult == null || score > maxResult.getScore()) {
                    MovingPawn pawn = action.getPawn();
                    maxResult = new MiniMaxResult(score, pawn.getRow(), pawn.getCol(), action.getMovement());
                }
            }
        }

        long endMillis = System.currentTimeMillis();
        System.out.printf("%d ms\n", (endMillis - startMillis));

        return maxResult;
    }

    private static int max(MovingBoard game, int depth, int alpha, int beta) {
        if (depth >= MAX_DEPTH) {
            return evaluate(game);
        }

        int maxScore = Integer.MIN_VALUE;

        for (Action action : getActions(game, true)) {
            MovingPawn pawn = action.getPawn();
            Pawn.PawnMovement movement = action.getMovement();

            game.move(pawn, movement);
            int score = min(game, depth + 1, Math.max(alpha, maxScore), beta);
            game.revertMove();

            if (score > maxScore) {
                maxScore = score;
            }

            if (maxScore >= beta) {
                return maxScore;
            }
        }

        return maxScore;
    }

    private static int min(MovingBoard game, int depth, int alpha, int beta) {
        if (depth >= MAX_DEPTH) {
            return evaluate(game);
        }

        int minScore = Integer.MAX_VALUE;

        for (Action action : getActions(game, false)) {
            MovingPawn pawn = action.getPawn();
            Pawn.PawnMovement movement = action.getMovement();

            game.move(pawn, movement);
            int score = max(game, depth + 1, alpha, Math.min(beta, minScore)) * -1;
            game.revertMove();

            if (score < minScore) {
                minScore = score;
            }

            if (minScore <= alpha) {
                return minScore;
            }
        }

        return minScore;
    }

    private static Collection<Action> getActions(MovingBoard game, boolean max) {
        List<Action> actions = new ArrayList<>();
        Collection<MovingPawn> pawns = max ? game.getMaxPawns() : game.getMinPawns();
        Pawn.PawnMovement[] movements = Pawn.PawnMovement.values();

//        Collection<MovingPawn> winningPawns = pawns.stream().filter(p -> p.getRow() + p.getDirection() == p.getPlayer().getGoal()).collect(Collectors.toList());
//        if (winningPawns.size() > 0) {
//            pawns = winningPawns;
//        }

        for (MovingPawn pawn : pawns) {
            for (Pawn.PawnMovement movement : movements) {
                if (pawn.isMoveValid(game.getBoard(), movement)) {
                    actions.add(new Action(pawn, movement));
                }
            }
        }

        Collections.shuffle(actions);
        return actions;
    }

    private static int evaluate(MovingBoard board) {
        return BoardEvaluator.evaluate(board);
    }

    static class Action {
        private final MovingPawn pawn;
        private final Pawn.PawnMovement movement;

        public Action(MovingPawn pawn, Pawn.PawnMovement movement) {
            this.pawn = pawn;
            this.movement = movement;
        }

        public MovingPawn getPawn() {
            return pawn;
        }

        public Pawn.PawnMovement getMovement() {
            return movement;
        }
    }
}
